import { WasmCrypto } from "cojson/crypto/WasmCrypto";
import { beforeEach, describe, expect, test, vi } from "vitest";
import {
  Account,
  AccountClass,
  AuthSecretStorage,
  CoMap,
  Group,
  InMemoryKVStore,
  JazzAuthContext,
  KvStoreContext,
  co,
} from "../exports";
import {
  JazzContextManager,
  JazzContextManagerAuthProps,
  JazzContextManagerBaseProps,
} from "../implementation/ContextManager";
import {
  createJazzContext,
  randomSessionProvider,
} from "../implementation/createContext";
import {
  createJazzTestAccount,
  getPeerConnectedToTestSyncServer,
  setupJazzTestSync,
} from "../testing";

const Crypto = await WasmCrypto.create();

class TestJazzContextManager<Acc extends Account> extends JazzContextManager<
  Acc,
  JazzContextManagerBaseProps<Acc> & {
    defaultProfileName?: string;
    AccountSchema?: AccountClass<Acc>;
  }
> {
  async createContext(
    props: JazzContextManagerBaseProps<Acc> & {
      defaultProfileName?: string;
      AccountSchema?: AccountClass<Acc>;
    },
    authProps?: JazzContextManagerAuthProps,
  ) {
    const context = await createJazzContext<Acc>({
      credentials: authProps?.credentials,
      defaultProfileName: props.defaultProfileName,
      newAccountProps: authProps?.newAccountProps,
      peersToLoadFrom: [getPeerConnectedToTestSyncServer()],
      crypto: Crypto,
      sessionProvider: randomSessionProvider,
      authSecretStorage: this.getAuthSecretStorage(),
      AccountSchema: props.AccountSchema,
    });

    this.updateContext(props, {
      me: context.account,
      node: context.node,
      done: () => {
        context.done();
      },
      logOut: async () => {
        await context.logOut();
      },
    });
  }
}

describe("ContextManager", () => {
  let manager: TestJazzContextManager<Account>;
  let authSecretStorage: AuthSecretStorage;

  function getCurrentValue() {
    return manager.getCurrentValue() as JazzAuthContext<Account>;
  }

  beforeEach(async () => {
    KvStoreContext.getInstance().initialize(new InMemoryKVStore());
    authSecretStorage = new AuthSecretStorage();
    await authSecretStorage.clear();
    await setupJazzTestSync();

    manager = new TestJazzContextManager<Account>();
  });

  test("creates new context when initialized", async () => {
    await manager.createContext({});

    const context = getCurrentValue();

    expect(context.me.profile?.name).toBe("Anonymous user");
    expect(context.node).toBeDefined();
    expect(manager.getCurrentValue()).toBeDefined();
  });

  test("creates new context when initialized with default profile name", async () => {
    await manager.createContext({
      defaultProfileName: "Test User",
    });

    const context = getCurrentValue();

    expect(context.me.profile?.name).toBe("Test User");
    expect(context.node).toBeDefined();
    expect(manager.getCurrentValue()).toBeDefined();
  });

  test("handles authentication with credentials", async () => {
    const account = await createJazzTestAccount();

    // First create an initial context to get credentials
    await manager.createContext({});

    const credentials = {
      accountID: account.id,
      accountSecret: account._raw.core.node.account.agentSecret,
      provider: "test",
    };

    // Authenticate with those credentials
    await manager.authenticate(credentials);

    expect(getCurrentValue().me.id).toBe(credentials.accountID);
  });

  test("calls onLogOut callback when logging out", async () => {
    const onLogOut = vi.fn();
    await manager.createContext({ onLogOut });

    await manager.logOut();

    expect(onLogOut).toHaveBeenCalled();
  });

  test("notifies listeners of context changes", async () => {
    const listener = vi.fn();
    manager.subscribe(listener);

    await manager.createContext({});

    expect(listener).toHaveBeenCalled();
  });

  test("cleans up context when done", async () => {
    await manager.createContext({});

    const context = manager.getCurrentValue();
    expect(context).toBeDefined();

    manager.done();

    // Should still have context, just cleaned up
    expect(manager.getCurrentValue()).toBe(context);
  });

  test("calls onAnonymousAccountDiscarded when authenticating from anonymous user", async () => {
    const onAnonymousAccountDiscarded = vi.fn();
    const account = await createJazzTestAccount();

    // Create initial anonymous context
    await manager.createContext({ onAnonymousAccountDiscarded });
    const anonymousAccount = getCurrentValue().me;

    // Authenticate with credentials
    await manager.authenticate({
      accountID: account.id,
      accountSecret: account._raw.core.node.account.agentSecret,
      provider: "test",
    });

    // Verify callback was called with the anonymous account
    expect(onAnonymousAccountDiscarded).toHaveBeenCalledWith(anonymousAccount);
  });

  test("does not call onAnonymousAccountDiscarded when authenticating from authenticated user", async () => {
    const onAnonymousAccountDiscarded = vi.fn();
    const account = await createJazzTestAccount();

    await manager.getAuthSecretStorage().set({
      accountID: account.id,
      accountSecret: account._raw.core.node.account.agentSecret,
      provider: "test",
    });

    // Create initial authenticated context
    await manager.createContext({ onAnonymousAccountDiscarded });

    // Authenticate with same credentials
    await manager.authenticate({
      accountID: account.id,
      accountSecret: account._raw.core.node.account.agentSecret,
      provider: "test",
    });

    // Verify callback was not called
    expect(onAnonymousAccountDiscarded).not.toHaveBeenCalled();
  });

  test("the migration should be applied correctly on existing accounts ", async () => {
    class AccountRoot extends CoMap {
      value = co.string;
    }

    let lastRootId: string | undefined;

    class CustomAccount extends Account {
      root = co.ref(AccountRoot);

      migrate() {
        this.root = AccountRoot.create({
          value: "Hello",
        });
        lastRootId = this.root.id;
      }
    }
    const customManager = new TestJazzContextManager<CustomAccount>();

    // Create initial anonymous context
    await customManager.createContext({
      AccountSchema: CustomAccount,
    });

    const account = (
      customManager.getCurrentValue() as JazzAuthContext<CustomAccount>
    ).me;

    await customManager.authenticate({
      accountID: account.id,
      accountSecret: account._raw.core.node.account.agentSecret,
      provider: "test",
    });

    const me = await CustomAccount.getMe().ensureLoaded({ root: {} });

    expect(me.root.id).toBe(lastRootId);
  });

  test("the migration should be applied correctly on existing accounts (2)", async () => {
    class AccountRoot extends CoMap {
      value = co.number;
    }

    class CustomAccount extends Account {
      root = co.ref(AccountRoot);

      async migrate(this: CustomAccount) {
        if (this.root === undefined) {
          this.root = AccountRoot.create({
            value: 1,
          });
        } else {
          const { root } = await this.ensureLoaded({ root: {} });

          root.value = 2;
        }
      }
    }
    const customManager = new TestJazzContextManager<CustomAccount>();

    // Create initial anonymous context
    await customManager.createContext({
      AccountSchema: CustomAccount,
    });

    const account = (
      customManager.getCurrentValue() as JazzAuthContext<CustomAccount>
    ).me;

    await customManager.authenticate({
      accountID: account.id,
      accountSecret: account._raw.core.node.account.agentSecret,
      provider: "test",
    });

    const me = await CustomAccount.getMe().ensureLoaded({ root: {} });

    expect(me.root.value).toBe(2);
  });

  test("onAnonymousAccountDiscarded should work on transfering data between accounts", async () => {
    class AccountRoot extends CoMap {
      value = co.string;
      transferredRoot = co.optional.ref(AccountRoot);
    }

    class CustomAccount extends Account {
      root = co.ref(AccountRoot);

      migrate() {
        if (this.root === undefined) {
          this.root = AccountRoot.create({
            value: "Hello",
          });
        }
      }
    }

    const onAnonymousAccountDiscarded = async (
      anonymousAccount: CustomAccount,
    ) => {
      const anonymousAccountWithRoot = await anonymousAccount.ensureLoaded({
        root: {},
      });

      const meWithRoot = await CustomAccount.getMe().ensureLoaded({ root: {} });

      const rootToTransfer = anonymousAccountWithRoot.root;

      await rootToTransfer._owner.castAs(Group).addMember(meWithRoot, "admin");

      meWithRoot.root.transferredRoot = rootToTransfer;
    };

    const customManager = new TestJazzContextManager<CustomAccount>();

    // Create initial anonymous context
    await customManager.createContext({
      onAnonymousAccountDiscarded,
      AccountSchema: CustomAccount,
    });

    const account = await createJazzTestAccount({
      isCurrentActiveAccount: true,
      AccountSchema: CustomAccount,
    });

    await customManager.authenticate({
      accountID: account.id,
      accountSecret: account._raw.core.node.account.agentSecret,
      provider: "test",
    });

    const me = await CustomAccount.getMe().ensureLoaded({ root: {} });

    expect(me.root.transferredRoot?.value).toBe("Hello");
  });
});
